PROGRAM FileDescEditor;
{$A+,B-,D-,E-,F-,G+,L+,N-,O-,R+,S+,V-,X-}
{$M 8192,0,655360}

(* ----------------------------------------------------------------------
   A Simple 4DOS File Description Editor

   (c) 1992, 1993 Copyright by

       David Frey,         & Tom Bowden
       Urdorferstrasse 30    1575 Canberra Drive
       8952 Schlieren ZH     Stone Mountain, GA 30088-3629
       Switzerland           USA

       Code created using Turbo Pascal 7.0, (c) Borland International 1992

   DISCLAIMER: This program is freeware: you are allowed to use, copy
               and change it free of charge, but you may not sell or hire
               4DESC. The copyright remains in our hands.

               If you make any (considerable) changes to the source code,
               please let us know. (send a copy or a listing).
               We would like to see what you have done.

               We, David Frey and Tom Bowden, the authors, provide absolutely
               no warranty of any kind. The user of this software takes the
               entire risk of damages, failures, data losses or other
               incidents.

   NOTES:      4DESC was modified extensively by Tom Bowden,
               August-October 1992.

               Among the changes:

               Screen layout now resembles the 4DOS SELECT screen.
               (The original screen had apparently been based on Larry
               Edwards' 4FILES).

               The display now is sorted.

               The program now is always in edit mode.

               Alt-T (cuT to buffer) now is Alt-M (Move to buffer).
               Alt-D now deletes a file description.
               Alt-X now exits the program.

               F1 now displays a help screen.
               F2 now changes drive.
               F3 now changes to the highlighted directory.
               F4 now changes to the parent directory.
               F10 now saves the current file descriptions.

               The screen colors were changed, and stored as CONST for
               easier maintenance.  VGA is no longer required.

               4DESC can now write and display file descriptions for
               directory entries.

               SaveDescriptions now strips trailing spaces from
               file extensions and leading and trailing spaces from
               file descriptions.

               4DESC does not presently write file descriptions longer
               than 40 characters.  When saving, any longer descriptions
               in the current directory will be truncated.  The user is
               warned when reading a directory having extended file
               descriptions.

               Handling of extended program information has not been tested.

   ADDITIONS TO TOM BOWDENS'S IMPROVEMENTS BY DAVID FREY:

               I have split 4DESC.PAS into units:
                StringDateHandling, DisplayKeyboardAndCursor, HandleINIFile

               Monochrome / Color display detection. /mono switch.

               Insert mode cursor is underline, overwrite is block cursor.
               (as in 4DOS)

               Since 4DOS 4.01 has introduced the `DescriptionMax' statement;
               references to fix description lengths have been removed.

               4DESC is now international: it chooses the appropriate date
               and time formats on startup. [by using DOS's function $38:
               Get/Set Country Data. DOS get the country information via
               COUNTRY= and COUNTRY.SYS].

               4TOOLS.INI file introduced. Colors and Time/Date formats can
               now be changed without recompiling 4DESC.PAS (for people
               without Turbo Pascal). 4DESC checks its startup directory,
               environment variable 4TOOLS and PATH to locate 4TOOLS.INI.

               "Change drive" will not change to drives which are not ready.

   A few new tweaks by Tom Bowden:

               "Change drive" will not change to drives which contain
               no files.

               New handling of command line parameters.  The /mono, /help,
               and selected directory params may be used together.  Note
               that the optional selected directory must be the last
               parameter entered.

               The status line now displays the 4DOS version (if running
               under 4DOS), and shows "Edit" and Cut" rather than "*"
               and "()".

   More additions by David Frey:

               Maximum number of files in a directory raised to 417
               descriptions. A warning ("Description file will be truncated")
               will appear if more than MaxDesc files are stored in a
               directory going to be edited with 4DESC.
               This prevents unintentional cutting of your description file.

               Yet another function key binding:
                F3 : View file (with list - whatever list may be
                                (internal 4DOS, external viewer))
                F4 : Change Dir
                F5 : Change to parent
                F6 : Change drive

               In 4TOOLS.INI the LeftJust variable has been added.

               Shelling out to 4DOS has been added (Alt-S or Shift-F10)

   More additions by Tom Bowden:

               In 4TOOLS.INI the FullSize and Viewer variables have
               been added.
               Get4DOSVer has been modified to display the correct
               minor version number and to check for NDOS.

   ADDITIONS TO TOM BOWDENS'S IMPROVEMENTS BY DAVID FREY:

   ||          Static description management has been replaced by a dynamic
   ||          one, i.e. the description data will be allocated on the heap.

               4DESC uses a TCollection (a Turbo Vision Object) to handle the
               descriptions. The description handling resides in DESCRIPT.PAS.

               Upper file limit (per directory) is now

                min { MaxCollectionSize = 16'380, available base memory }


               A Delimiters-variable has been added, specifying legal
               separators between words (used by Ctrl-Left / Right = Move
               a word to the left / right)

   ----------------------------------------------------------------------- *)

USES {$IFOPT G+} Test286, {$ENDIF}
     Crt, Dos,
     Memory,
     StringDateHandling, DisplayKeyboardAndCursor, HandleINIFile,
     DescriptionHandling;

CONST DelimiterTable : STRING = ',.();:-!?/[]{}+*=''`"@%&$_';

VAR  EdStart     : BYTE;
     DescLen     : BYTE;

     ActDir      : DirStr;
     StartDir    : DirStr;

     StartIndex  : INTEGER;
     Index       : INTEGER;

     CutPasteDesc: DescStr;
     Changed     : BOOLEAN;
     IORes       : INTEGER;

     NewDir      : DirStr;
     NewName     : NameStr;
     NewExt      : ExtStr;

     FirstParam  : STRING[2];
     i           : BYTE;
     DoShowHelp  : BOOLEAN;

(*-------------------------------------------------------- Display-Routines *)
PROCEDURE WriteFileEntry(Index: INTEGER; Hilighted: BOOLEAN);

VAR FileEntry : PFileData;

BEGIN
 GotoXY(1,3+Index-StartIndex);
 IF (Index >= 0) AND (Index < FileList^.Count) THEN
  BEGIN
   FileEntry := NILCheck(FileList^.At(Index));

   IF Hilighted THEN
    BEGIN TextColor(SelectFg); TextBackGround(SelectBg); END
   ELSE
    BEGIN
     TextBackGround(NormBg);

     IF FileEntry^.GetSize <> DirSize THEN TextColor(NormFg)
                                      ELSE TextColor(DirFg)
    END;
   Write(FileEntry^.FormatDescription);
   IF Length(FileEntry^.GetDesc) < DescLen THEN ClrEol;
  END
 ELSE ClrEol;
END;  (* WriteFileEntry *)

PROCEDURE DrawDirLine;

BEGIN
 GetDir(0,ActDir);
 IF ActDir[Length(ActDir)] <> '\' THEN ActDir := ActDir + '\';
 UpString(ActDir);
 TextColor(DirFg); TextBackGround(NormBg);
 GotoXY(1,2); Write(' ',ActDir); ClrEol;
END; (* DrawDirLine *)

PROCEDURE ReDrawScreen;

VAR Index: INTEGER;

BEGIN
 GetDir(0,ActDir);
 FOR Index := StartIndex TO StartIndex+MaxLines-4 DO
  WriteFileEntry(Index,FALSE);
END; (* ReDrawScreen *)


(*-------------------------------------------------------- Read-Directory *)
PROCEDURE ReadFiles;

VAR i              : BYTE;
    ch             : WORD;
    Dir            : PathStr;

BEGIN
 Changed := FALSE; DescLong := FALSE;
 Index   := 0; StartIndex := 0;
 Dir := FExpand('.');

 IF FileList <> NIL THEN
  BEGIN
   Dispose(FileList,Done); FileList := NIL;
  END;

 TextColor(StatusFg); TextBackGround(StatusBg);
 GotoXY(1,MaxLines); Write('Scanning directory `',Dir,'''.....  please wait.'); ClrEol;

 FileList := NIL; FileList := New(PFileList,Init(Dir));
 IF FileList = NIL THEN Abort('Unable to allocate FileList');

 IF (FileList^.Status = ListTooManyFiles) OR
    (FileList^.Status = ListOutofMem) THEN
  BEGIN
   TextColor(NormFg); TextBackGround(NormBg);
   FOR i := 3 TO MaxLines-1 DO
    BEGIN
     GotoXY(1,i); ClrEol;
    END;
   IF FileList^.Status = ListTooManyFiles THEN
    ReportError('Warning! Too many files in directory, description file will be truncated! (Key)',(CutPasteDesc <> ''),Changed)
   ELSE
    ReportError('Warning! Out of memory, description file will be truncated! (Key)',(CutPasteDesc <> ''),Changed);
  END;

 IF FileList^.Count > 0 THEN
  BEGIN
   DrawMainScreen(Index,FileList^.Count);
   DrawDirLine;
  END;

 IF DescLong THEN
  BEGIN
   TextColor(NormFg); TextBackGround(NormBg);
   FOR i := 3 TO MaxLines-1 DO
    BEGIN
     GotoXY(1,i); ClrEol;
    END;
   ReportError('Warning! Some descriptions are too long; they will be truncated. Press any key.',(CutPasteDesc <> ''),Changed);
  END;
END;  (* ReadFiles *)

(*-------------------------------------------------------- Save Descriptions *)
PROCEDURE SaveDescriptions;

VAR DescFile : TEXT;
    DescSaved: BOOLEAN;
    Time     : DateTime;
    ch       : WORD;
    FileEntry: PFileData;
    

 PROCEDURE SaveEntry(FileEntry: PFileData); FAR;

 VAR Desc     : DescStr;
     ProgInfo : STRING;
     Dir      : DirStr;
     BaseName : NameStr;
     Ext      : ExtStr;

 BEGIN
  Desc := FileEntry^.GetDesc;
  IF Desc <> '' THEN
   BEGIN
    FSplit(FileEntry^.GetName,Dir,Basename,Ext);
    StripTrailingSpaces(BaseName);
    Write(DescFile,BaseName);

    StripLeadingSpaces(Ext);
    StripTrailingSpaces(Ext);
    IF Ext <> '' THEN Write(DescFile,Ext);

    StripLeadingSpaces(Desc); StripTrailingSpaces(Desc);
    Write(DescFile,' ',Desc);

    ProgInfo :=  FileEntry^.GetProgInfo;
    IF ProgInfo <> '' THEN Write(DescFile,#4,ProgInfo);
    WriteLn(DescFile);
   END;
 END; (* SaveEntry *)

BEGIN
 DescSaved := FALSE;
 IF DiskFree(0) < FileList^.Count*SizeOf(TFileData) THEN
  ReportError('Probably out of disk space. Nevertheless trying to save DESCRIPT.ION...',(CutPasteDesc <> ''),Changed);

 TextColor(StatusFg); TextBackGround(StatusBg);
 GotoXY(1,MaxLines); Write('Saving descriptions........  please wait.'); ClrEol;

 Assign(DescFile,'DESCRIPT.ION'); SetFAttr(DescFile,Archive);
 Rewrite(DescFile);
 IF IOResult > 0 THEN ReportError('Unable to write DESCRIPT.ION !',(CutPasteDesc <> ''),Changed)
 ELSE
  BEGIN FileList^.ForEach(@SaveEntry); DescSaved := TRUE; END;
 {$I-}
 Close(DescFile);
 {$I+}

 IF IOResult > 0 THEN ReportError('Unable to write DESCRIPT.ION !',(CutPasteDesc <> ''),Changed)
 ELSE
  BEGIN
   IF DescSaved THEN SetFAttr(DescFile, Archive + Hidden)
                ELSE Erase(DescFile);  (* Don't keep zero-byte file. *)
   Changed := FALSE; DrawStatusLine(TRUE,(CutPasteDesc <> ''),Changed);
  END;
END;  (* SaveDescriptions *)

(*-------------------------------------------------------- Edit Descriptions *)
PROCEDURE EditDescriptions;

CONST kbLeft     = $4B00;   kbRight    = $4D00;
      kbUp       = $4800;   kbDown     = $5000;
      kbHome     = $4700;   kbEnd      = $4F00;
      kbPgUp     = $4900;   kbPgDn     = $5100;
      kbCtrlLeft = $7300;   kbCtrlRight= $7400;
      kbCtrlPgDn = $7600;   kbCtrlPgUp = $8400;
      kbCtrlHome = $7700;   kbCtrlEnd  = $7500;
      kbEnter    = $0D;     kbEsc      = $1B;

      kbIns      = $5200;   kbDel      = $5300;
      kbBack     = $08;

      kbGrayMinus= $4A2D;   kbGrayPlus = $4E2B;

      kbAltC     = $2E00;   kbAltP     = $1900;
      kbAltD     = $2000;   kbAltL     = $2600;
      kbAltM     = $3200;   kbAltT     = $1400;
      kbAltS     = $1F00;   kbAltV     = $2F00;
      kbAltX     = $2D00;

      kbF1       = $3B00;   kbF2       = $3C00;
      kbF3       = $3D00;   kbF4       = $3E00;
      kbF5       = $3F00;   kbF6       = $4000;
      kbF10      = $4400;   kbShiftF10 = $5D00;

VAR Key          : WORD;
    Drv          : STRING[3];
    LastDrv      : CHAR;
    x,y,l        : BYTE;
    EditStr      : DescStr;
    Overwrite    : BOOLEAN;
    Cursor       : WORD;
    OldDir       : DirStr;
    ActFileData  : PFileData;
    n            : NameExtStr;

 PROCEDURE UpdateLineNum(Index: INTEGER);

 BEGIN
  WriteFileEntry(Index,TRUE);
  TextColor(StatusFg); TextBackGround(StatusBg);
  GotoXY(66,1); Write(Index+1:5);

  IF Changed THEN DrawStatusLine(FALSE,(CutPasteDesc <> ''),Changed);

  IF Index < FileList^.Count THEN
   BEGIN
    x := 1;
    y := 3+Index-StartIndex;
    GotoXY(EdStart,y);
    TextColor(SelectFg); TextBackGround(SelectBg);
    EditStr := PFileData(FileList^.At(Index))^.GetDesc;
    Write(EditStr);
    IF Length(EditStr) < DescLen THEN ClrEol;
    GotoXY(EdStart+x-1,y);
   END;

  ActFileData := NILCheck(FileList^.At(Index));
 END;

 PROCEDURE PrevIndex(VAR Index: INTEGER);

 BEGIN
  Index := Max(Index-1,0);
  IF Index <= StartIndex THEN
   BEGIN
    StartIndex := Max(Index-ScreenSize,0);
    RedrawScreen;
   END;
  UpdateLineNum(Index);
 END; (* NextIndex *)

 PROCEDURE NextIndex(VAR Index: INTEGER);

 BEGIN
  Index := Min(Index+1,FileList^.Count-1);
  IF Index > StartIndex+ScreenSize THEN
   BEGIN
    StartIndex := Index-ScreenSize;
    RedrawScreen;
   END;
  UpdateLineNum(Index);
 END; (* NextIndex *)

 PROCEDURE QuerySaveDescriptions;

 VAR ch: CHAR;

 BEGIN
  TextColor(StatusFg); TextBackGround(StatusBg);
  IF Changed THEN
   BEGIN
    REPEAT
     GotoXY(1,MaxLines);
     Write(Chars(' ',11));
     Write('Descriptions have been edited. Shall they be saved (Y/N) ?');
     ClrEol;
     ch := UpCase(ReadKey);
    UNTIL (ch = 'Y') OR (ch = 'N');
    IF ch = 'Y' THEN SaveDescriptions;
   END;
 END; (* QuerySaveDescriptions *)

 PROCEDURE DirUp;

 BEGIN
  IF Changed THEN QuerySaveDescriptions;
  ChDir('..');
  IF IOResult = 0 THEN
   BEGIN
    ReadFiles;
    RedrawScreen;

    DrawStatusLine(FALSE,(CutPasteDesc <> ''),Changed);
    Index := 0; UpdateLineNum(Index);
   END;
 END;  (* DirUp *)

 PROCEDURE DirDown;

 BEGIN
  IF (Index < FileList^.Count) THEN
   BEGIN
    n  := ActFileData^.GetName;
    IF (ActFileData^.GetSize = DirSize) AND (n[1] <> '.') THEN
     BEGIN
      IF Changed THEN QuerySaveDescriptions;
      ChDir(n);
      IF IOResult = 0 THEN
       BEGIN
        ReadFiles;
        RedrawScreen;
       END;
      DrawStatusLine(FALSE,(CutPasteDesc <> ''),Changed);
      Index := 0; UpdateLineNum(Index);
    END;  (* IF Description[Index].Size = DirSize *)
   END;
 END;  (* DirDown *)

 FUNCTION IsADelimiter(c: CHAR): BOOLEAN;
 (* used by Ctrl-Left resp Ctrl-Right *)

 BEGIN
  IsADelimiter := (Pos(c,DelimiterTable) > 0);
 END;

BEGIN  (* EditDescriptions *)
 Index := 0; UpdateLineNum(Index);

 Overwrite := FALSE; ResetCursor(Overwrite);
 EditStr := ActFileData^.GetDesc;

 REPEAT
  Key := GetKey;
  CASE Key OF
   kbUp       : BEGIN
                 ActFileData^.AssignDesc(EditStr); WriteFileEntry(Index,FALSE);
                 PrevIndex(Index); 
                END; (* Up *)

   kbDown     : BEGIN
                 ActFileData^.AssignDesc(EditStr); WriteFileEntry(Index,FALSE);
                 NextIndex(Index);
                END; (* Down *)

   kbLeft     : BEGIN
                 x := Max(1,x-1);
                 GotoXY(EdStart+x-1,y);
                END; (* Left *)

   kbRight    : BEGIN
                 IF (x <= Length(EditStr)) AND (x < DescLen) THEN INC(x);
                 GotoXY(EdStart+x-1,y);
                END; (* Right *)

   kbCtrlLeft : BEGIN
                 DEC(x);
                 WHILE (x > 0) AND IsADelimiter(EditStr[x]) DO DEC(x);
                 WHILE (x > 0) AND NOT IsADelimiter(EditStr[x]) DO DEC(x);
                 INC(x); GotoXY(EdStart+x-1,y);
                END;

   kbCtrlRight: BEGIN
                 l := Length(EditStr);
                 WHILE (x < l) AND NOT IsADelimiter(EditStr[x]) DO INC(x);
                 WHILE (x < l) AND IsADelimiter(EditStr[x]) DO INC(x);
                 GotoXY(EdStart+x-1,y);
                END;

   kbHome     : BEGIN
                 x := 1;
                 GotoXY(EdStart+x-1,y);
                END; (* Home *)

   kbEnd      : BEGIN
                 x := Length(EditStr);
                 IF x < DescLen THEN INC(x);
                 GotoXY(EdStart+x-1,y);
                 END; (* End *)

   kbCtrlEnd  : BEGIN
                 Delete(EditStr,x,DescLen);
                 ActFileData^.AssignDesc(EditStr);
                 WriteFileEntry(Index,TRUE);

                 Changed := TRUE;
                 DrawStatusLine(FALSE,(CutPasteDesc <> ''),Changed);
                END;  (* ^End *)

   kbIns      : BEGIN
                 Overwrite := NOT Overwrite;
                 ResetCursor(Overwrite);
                END; (* Ins *)

   kbDel      : BEGIN
                 IF x <= Length(EditStr) THEN Delete(EditStr,x,1);
                 ActFileData^.AssignDesc(EditStr);
                 WriteFileEntry(Index,TRUE);

                 Changed := TRUE;
                 DrawStatusLine(FALSE,(CutPasteDesc <> ''),Changed);

                 GotoXY(EdStart+x-1,y);
                END; (* Del *)

   kbBack     : BEGIN
                 Delete(EditStr,x-1,1);
                 ActFileData^.AssignDesc(EditStr);
                 IF x > 1 THEN
                  BEGIN
                   DEC(x);
                   IF x > Length(EditStr) THEN x := Length(EditStr)+1;
                  END;
                 WriteFileEntry(Index,TRUE);

                 Changed := TRUE;
                 DrawStatusLine(FALSE,(CutPasteDesc <> ''),Changed);
                 GotoXY(EdStart+x-1,y);
                END; (* Back *)

   kbPgUp     : BEGIN
                 ActFileData^.AssignDesc(EditStr);
                 WriteFileEntry(Index,FALSE);
                 
                 Index := Max(Index-ScreenSize,0);
                 StartIndex := Index;
                 RedrawScreen;
                 UpdateLineNum(Index);
                END; (* PgUp *)

   kbPgDn     : BEGIN
                 ActFileData^.AssignDesc(EditStr); 
                 Index := Min(Index+ScreenSize,FileList^.Count-1);
                 StartIndex := Max(Index-ScreenSize,0);
                 WriteFileEntry(Index,FALSE);

                 RedrawScreen;
                 UpdateLineNum(Index);
                END; (* PgDn *)

   kbCtrlPgUp : BEGIN
                 ActFileData^.AssignDesc(EditStr);
                 WriteFileEntry(Index,FALSE);

                 StartIndex := 0; Index := 0;
                 RedrawScreen;
                 UpdateLineNum(Index);
                END; (* ^PgUp *)

   kbCtrlPgDn : BEGIN
                 ActFileData^.AssignDesc(EditStr);
                 WriteFileEntry(Index,FALSE);

                 StartIndex := Max(FileList^.Count-ScreenSize,0);
                 Index := FileList^.Count-1;
                 RedrawScreen;
                 UpdateLineNum(Index);
                END; (* ^PgDn *)

   kbAltD     : BEGIN
                 ActFileData^.AssignDesc('');

                 Changed := TRUE;
                 DrawStatusLine(FALSE,(CutPasteDesc <> ''),Changed);
                 WriteFileEntry(Index,FALSE);
                 NextIndex(Index);
                END; (* Alt-D *)

   kbAltM,
   kbAltT     : BEGIN
                 CutPasteDesc := ActFileData^.GetDesc;
                 ActFileData^.AssignDesc(''); EditStr := '';

                 Changed := TRUE;
                 DrawStatusLine(FALSE,(CutPasteDesc <> ''),Changed);
                 WriteFileEntry(Index,FALSE);
                 NextIndex(Index);
                END; (* Alt-M / Alt-T *)

   kbAltC     : BEGIN
                 CutPasteDesc := ActFileData^.GetDesc;

                 DrawStatusLine(FALSE,(CutPasteDesc <> ''),Changed);
                 WriteFileEntry(Index,TRUE);
                END; (* Alt-C *)

   kbAltP     : IF CutPasteDesc > '' THEN
                 BEGIN
                  ActFileData^.AssignDesc(CutPasteDesc);
                  WriteFileEntry(Index,FALSE);

                  Changed := TRUE;
                  DrawStatusLine(FALSE,(CutPasteDesc <> ''),Changed);
                  NextIndex(Index);
                 END; (* Alt-P *)

   kbF1       : BEGIN                                   (* F1: Help *)
                 ShowHelp;
                 ResetCursor(Overwrite);
                 DrawMainScreen(Index,FileList^.Count);
                 DrawDirLine;
                 RedrawScreen;
                 UpdateLineNum(Index);
                END;  (* F1 *)

   kbAltL,
   kbF6       : BEGIN                                   (* F6: Change Drive *)
                 IF Changed THEN QuerySaveDescriptions;

                 ASM
                  mov ah,0eh       (* Select Disk *)
                  mov dl,3
                  int 21h
                  add al,'@'
                  mov LastDrv,al
                 END;

                 (* Tom's solution has been commented out, Dave.
                    Reason: LastDrive has not been detected correctly. *)

                 IF LastDrv > 'Z' THEN LastDrv := 'Z';

                 TextColor(StatusFg); TextBackGround(StatusBg); Drv := ' :';
                 REPEAT
                  GotoXY(1,MaxLines);
                  Write(' New drive letter (A..',LastDrv,'): ');
                  ClrEol;
                  Drv[1] := UpCase(ReadKey);
                 UNTIL (Drv[1] >= 'A') AND (Drv[1] <= LastDrv);
                 IF Drv[1] <= 'B' THEN Drv := Drv + '\';
                 OldDir := ActDir;
                 ChDir(Drv); IORes := IOResult;
                 IF IORes = 0 THEN
                  BEGIN
                   GetDir(0,ActDir); IORes := IOResult;
                   ReadFiles;
                   IF FileList^.Count = 0 THEN
                    BEGIN
                     IF (Length(OldDir) > 3) AND (OldDir[Length(OldDir)] = '\') THEN
                       Delete(OldDir,Length(OldDir),1);
                     ChDir(OldDir);
                     ReportError('There are no files on drive '+Drv+'. Press any key.',(CutPasteDesc <> ''),Changed);
                     ReadFiles;
                    END;
                   RedrawScreen;
                   DrawStatusLine(FALSE,(CutPasteDesc <> ''),Changed);;
                   Index := 0;
                   UpdateLineNum(Index);
                  END
                 ELSE
                  ReportError('Drive '+Drv+' not ready! Drive remains unchanged, press a key.',(CutPasteDesc <> ''),Changed);
                END;  (* F6 *)

   kbF4       : DirDown; (* F4 *)
   kbF5       : DirUp;   (* F5 *)

   kbEnter    : BEGIN
                 ActFileData^.AssignDesc(EditStr);
                 WriteFileEntry(Index,TRUE);
                 IF (Index < FileList^.Count) THEN
                  BEGIN
                   n  := ActFileData^.GetName;

                   IF (ActFileData^.GetSize = DirSize) THEN
                    IF (n[1] = '.') AND (n[2] = '.') THEN DirUp
                    ELSE
                    IF n[1] <> '.' THEN DirDown;
                  END;
                END; (* Enter *)
   kbF10,
   kbF2      : BEGIN                                   (* F10: Save *)
                SaveDescriptions;
                UpdateLineNum(Index);
               END; (* F10 or F2 *)
   kbAltS,
   kbShiftF10: BEGIN                                   (* Shell to 4DOS *)
                IF Changed THEN QuerySaveDescriptions;

                DoneMemory;
                SetMemTop(HeapPtr);

                NormVideo; ClrScr;
                WriteLn('Type `Exit'' to return to 4DESC.');
                SwapVectors;
                Exec(GetEnv('COMSPEC'),'');
                SwapVectors;

                SetMemTop(HeapEnd);
                InitMemory;
                ClrScr;
                DrawMainScreen(Index,FileList^.Count);
                DrawStatusLine(TRUE,(CutPasteDesc <> ''),Changed);
                DrawDirLine;
                IF DosError > 0 THEN ReportError('Can''t load command interpreter / program execution failed.',
                                                 (CutPasteDesc <> ''),Changed);;
                ReadFiles;
                RedrawScreen;
                UpdateLineNum(Index);
               END;
   kbAltV,
   kbF3      : IF (Index < FileList^.Count) THEN
                BEGIN
                 IF ActFileData^.GetSize <> DirSize THEN
                  BEGIN                                  (* F3: View File *)
                   FSplit(ActFileData^.GetName,NewDir,NewName,NewExt);
                   StripTrailingSpaces(NewName);

                   DoneMemory;
                   SetMemTop(HeapPtr);

                   SwapVectors;
                   Exec(GetEnv('COMSPEC'),'/c '+ListCmd+' '+ActDir+'\'+NewName+NewExt);
                   SwapVectors;

                   SetMemTop(HeapEnd);
                   InitMemory;

                   ClrScr;
                   DrawMainScreen(Index,FileList^.Count);
                   DrawStatusLine(TRUE,(CutPasteDesc <> ''),Changed);
                   DrawDirLine;
                   IF DosError > 0 THEN ReportError('Can''t load command interpreter/program execution failed.',
                                                   (CutPasteDesc <> ''),Changed);
                   RedrawScreen;
                   UpdateLineNum(Index);
                 END;
                END;
  ELSE
   IF (Ord(Key) > 31) AND (Ord(Key) < 256) THEN
    BEGIN
     Changed := TRUE;
     DrawStatusLine(FALSE,(CutPasteDesc <> ''),Changed);

     IF Overwrite AND (x <= Length(EditStr)) THEN
      EditStr[x] := Chr(Key)
     ELSE
      EditStr := Copy(EditStr,1,x-1)+Chr(Key)+Copy(EditStr,x,255);
     INC(x);
     IF x > DescLen THEN x := DescLen;

     ActFileData^.AssignDesc(EditStr); WriteFileEntry(Index,TRUE);
     GotoXY(EdStart+x-1,y);
    END; (* all others *)

  END;  (* case *)
 UNTIL (Key = kbEsc) OR (Key = kbAltX);

 IF Changed THEN QuerySaveDescriptions;
END; (* EditDescriptions *)

(*-------------------------------------------------------- Main *)
BEGIN
 EdStart := 25+Length(DateFormat)+Length(TimeFormat);
 DescLen := ScreenWidth-EdStart;
 GetDir(0,StartDir); IORes := IOResult; DoShowHelp := FALSE;
 IF ParamCount > 0 THEN
  BEGIN
   FOR i := 1 TO Min(2,ParamCount) DO
    BEGIN
     FirstParam := ParamStr(i);
     IF (FirstParam[1] = '/') OR (FirstParam[1] = '-') THEN
      BEGIN
       IF NOT Monochrome THEN Monochrome := (UpCase(FirstParam[2]) = 'M');
       IF NOT DoShowHelp THEN DoShowHelp := (UpCase(FirstParam[2]) = 'H') OR
                                            (FirstParam[2] = '?');
      END;
    END;  (* for ... do begin *)
   FSplit(ParamStr(ParamCount), NewDir, NewName, NewExt);
   IF NewDir[Length(NewDir)] = '\' THEN NewDir[Length(NewDir)] := ' ';
   ChDir(NewDir);
  END;  (* if paramcount > 0 *)
 IORes := IOResult;
 Changed := FALSE; CutPasteDesc := '';
 ChooseColors(Monochrome);
 DrawMainScreen(0,0);
 IF INIFileExists THEN
  DelimiterTable := ReadSettingsString('misc','delimiters',DelimiterTable);
 DelimiterTable := ' '+DelimiterTable;

 IF DoShowHelp THEN ShowHelp;

 InitMemory;
 ReadFiles;
 RedrawScreen;
 EditDescriptions;
 Dispose(FileList,Done); FileList := NIL;
 DoneMemory;

 ChDir(StartDir);
 SetCursorShape(OrigCursor);
 NormVideo;
 ClrScr;
 WriteLn(Header);
 WriteLn;
 WriteLn('This program is freeware: you are allowed to use, copy it free');
 WriteLn('of charge, but you may not sell or hire 4DESC.');
END.
