/*  This is the source code of a utility "NEWZIP" that will re-compress files
    that were originally compressed with a PKZIP utility that was either prior
    to version 2 or not "Deflated" with maximum compression.

    What this program does is check to see if a user given file spec are ZIP
    files and if it is then it checks to see if the files are "Maximum
    Deflated".  Upon finding such files it first calls PKUNZIP to restore
    everything in each ZIP file and then it calls PKZIP to put the extracted
    files back into each ZIP file with maximum compression.

    The source code was written by George Heitz (CompuServe ID 76515,2432)
    and compiled with the Borland C++ ver 3.1 compiler.

    21 April 1993							    */

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <string.h>
#include <dir.h>
#include <ctype.h>

/* The following is the central directory structure type as defined in
   PKWare's APPNOTE.TXT for PKZIP 2.04					    */

struct central_dir
{
    long signature;
     int v_madeby;
     int v_needed;
     int gp_flag;
     int comp_mthd;
     int f_date;
     int f_time;
    long crc;
    long comp_size;
    long full_size;
     int fname_len;
     int xfield_len;
     int fcom_len;
     int dskno_strt;
     int in_attr;
    long ext_attr;
    long file_offs;
};

/* The following is the "End of central directory record" structure type as
   defined in PKWare's APPNOTE.TXT for PKZIP 2.04			    */

struct end_c_dir
{
    long signature;
     int disk_no;
     int c_dir_disk_no;
     int c_dir_ent_loc;
     int c_dir_entries;
    long c_dir_size;
    long c_dir_offset;
     int comment_len;
};
int check_zip(FILE *stream, char ch[], long *fptr);
void combine(char *str1);
int nomatch(char ch[]);
int do_rezip(char *ch);
void combine2(char *str1, char *str2);
void combine3(char *str1, char *str2);

int main(int argc, char *argv[])
{
   FILE *stream, *lfile;
   int i, j, k, done, code, log_flag;
   char filename[64], path[64], pathname[64], dos_call[64], ch[4], *ptr;
   char fn_done[3][13], logfile[64];
   struct ffblk ffblk;
   long fptr;
   strcpy(logfile,argv[0]);
   ptr = strrchr(logfile,'\\');
   if (argc <= 1)
   {
	printf("Usage: %s zipfile(.ZIP|.ext) [/L]\n",ptr + 1);
	return 1;
   }

/* The following retrieves the ZIP file spec from the command line	    */

   for (k = 1; k < argc; k++)
	if (argv[k][0] != '/' && argv[k][0] != '-')
	{
	    strcpy(filename,argv[k]);
	    break;
	}

/* The follow checks to see if there is an "L" switch to order a logfile to
   kept.  The log file, if it is kept is kept in the same path as NEWZIP.   */

   log_flag = 0;
   for (k = 1; k < argc; k++)
	if ((argv[k][0] == '/' || argv[k][0] == '-') &&
		toupper(argv[k][1]) == 'L')
	{
	    *(ptr+1) = 0;
	    strcat(logfile,"NEWZIP.LOG");
	    log_flag = 1;
	    break;
	}
   combine(filename);
   strupr(filename);
   strcpy(path,filename);
   ptr = strrchr(path,'\\');		/* Put path of ZIP filespec aside   */
   if (ptr != NULL)
	*(ptr+1) = 0;
   else
	*path = 0;
   done = findfirst(filename,&ffblk,0);
   if(done)
   {
       printf("Can't find any file of spec -- %s",filename);
       return 1;
   }
   i = 0;	/* counter for first three files that are rezipped	    */
   while(!done)	/* "while" loop to check all files of inputted file spec    */
   {
	strcpy(pathname,path);	/* add ZIP filespec path to found filename  */
	strcat(pathname,ffblk.ff_name);
	if ((stream = fopen(pathname, "rb")) == NULL)
	{
	    printf("Cannot open ZIP file, %s.\n",filename);
	    return 1;
	}
	code = 0;		/* code = 0 means file not max compressed   */

/* The following "for loop" is needed to check if a file from the beginning
  of the directory that has already been rezipped has been placed at the
  end of the directory by PKZIP						    */

	for (j = 0; j < i; j++)
	    if (strcmp(fn_done[j],ffblk.ff_name) == 0)
		code = 2;

/* Now if it has not already been rezipped check to see if it is first a ZIP
   file, and if not, see if it is already at max compression.		    */

	if (code == 0)
	    code = check_zip(stream,pathname,&ffblk.ff_fsize);
	fclose(stream);			/* close the potential ZIP file	    */
	if (!code)
	{
	    if (do_rezip(pathname))
		return 1;
	    if (log_flag == 1)
	    {
		log_flag = 2;
		lfile = fopen(logfile,"at");
	    }
	    if (i < 3)
		strcpy(fn_done[i++],ffblk.ff_name);

/* reset the file date and time of the ZIP file				    */

	    stream = fopen(pathname, "rb");
	    _dos_setftime(fileno(stream), ffblk.ff_fdate, ffblk.ff_ftime);
	    if (log_flag)
	    {
		fseek(stream,0L,SEEK_END);
		fptr = ftell(stream);	/* find length of rezipped file	    */

/* put entry into NEWZIP.LOG and do error checking			    */

		if (fprintf(lfile,"%12s  before -- %7ld, after -- %7ld, "
			"bytes saved -- %5ld\n", ffblk.ff_name,
			 ffblk.ff_fsize, fptr, ffblk.ff_fsize-fptr) == EOF)
		{
		    fclose(lfile);
		    log_flag = 0;
		}
	    }
	    fclose(stream);		       /* close rezipped file	    */
	}
	else if (code == 1)                    /* display error and continue*/
	    printf("%12s is already at maximum "
		"ZIP compression.\n",ffblk.ff_name);
	done = findnext(&ffblk);
   }
   return 0;
}


/* The following function checks first to see if the file that is to be
   rezipped is actually a ZIP file and then if it is a ZIP file it checks
   the compression method of the compressed files in the ZIP file to see
   if they are compressed with the "Maximum Deflation"			    */

int check_zip(FILE *stream, char filename[], long *flen)
{
   struct central_dir cd;
   struct end_c_dir ecd;
   char ch[4];
   int i, j, l, count, code;
   long k;
   fseek(stream,-22L,SEEK_END);
   fread(&ecd, sizeof(ecd), 1, stream);

/* If the file is a ZIP file with no zipfile comment then the "End of central
   dir record" structure will begin at 22 bytes from the end of the file.
   If it doesn't begin there then a search will be done for the signature   */

   if (ecd.signature != 0x06054b50)
   {					/* begin search for signature	    */
	j = 0;				/* j = 1 denotes signature found    */
	for (k = 534; k <= *flen; k += 512)
	{
	    fseek(stream,-k,SEEK_END);
	    i = 0;
	    for (l = 0; l < 512; l++)
	    {
		if ((ch[i] = fgetc(stream)) == 'P') /* first look for 'P'   */
		{
		    for (i = 1; i < 4; i++)
			if ((ch[i] = fgetc(stream)) == 'P')
			{
			    ch[0] = ch[i];
			    i = 0;
			}
		    if (nomatch(ch) == 0)   /* check for the rest of the    */
		    {                       /* bytes in the signature	    */
			j = 1;              /* if found then break to read  */
			break;		    /* the first central directory  */
		    }                       /* record			    */
		    else
			i = 0;		    /* if not found then continue   */
		}                           /* search			    */
	    }
	    if (j)
		break;			    /* break if signature found	    */
	    j = 0;
	}
	if (j == 0)
	{                                   /* after scan of entire file    */
					    /* and signature not found then */
					    /* display an error and return  */
	    printf("%s is not a ZIP file\n",filename);
	    return 2;
	}
	fread(&ecd.disk_no, sizeof(ecd)-4, 1, stream);
   }
   fseek(stream,ecd.c_dir_offset,SEEK_SET);
   code = 1;
   for (i = 0; i < ecd.c_dir_entries; i++)
   {
	fread(&cd,sizeof(cd),1,stream);

/* If central directory signature is not at the point pointed to by the "End
   of central directory" structure then the file is still not a ZIP file    */

	if (cd.signature != 0x02014b50)
	{
	    printf("%s is not a ZIP file.",filename);   /* display error    */
	    return 2;                                   /* and return	    */
	}
/* the following "if" statement check to see if compression method is prior
   to deflating and if so sets the "code" flag to zero			    */

	if (cd.comp_mthd > 0 && cd.comp_mthd < 8)
	    code = 0;

/* If the compression method is "0" then the file wasn't compressed.  The
   following "if" statement checks to see if the file is just "stored" and
   if it is then it check to see if it wasn't compressed because it was too
   small to waste time to be compressed.				    */

	if (cd.comp_mthd == 0 && cd.full_size > 256)
	    code = 0;

/* The following "if" statement checks that if the file was deflated by what
   level of deflation was it compressed.d				    */

	if (cd.comp_mthd == 8 && (cd.gp_flag/2 % 4) != 1)
	    code = 0;

/* Now we go on to check the next Central Directory entry		    */

	count = cd.xfield_len + cd.fcom_len + cd.fname_len;
	if (count > 0)
	    fseek(stream,count,SEEK_CUR);
   }
   return code;
}

/*  The following function checks to see if there was an extension put on to
    the ZIP file spec.  If not then it concatenates the ".ZIP" extension    */

void combine(char *str1)
{
    int i, j;
    i = strlen(str1) - 1;
    for (i = i; i > 0; i--)
    {
       if (str1[i] == '.')
	  return;
       if (str1[i] == '\\')
	  break;
    }
    strcat(str1,".ZIP");

}

/* This function just checks to see if the "End of central directory
   signature has been found.						    */

int nomatch(char ch[])
{
    int i;
    char test[4] = {'P','K',5,6};
    for (i = 0; i < 4; i++)
	if(ch[i] != test[i])
	    return 1;
    return 0;
}

/* This is the function that does the DOS calls to PKZIP and PKUNZIP.  Those
   executeable files are assumed to be in the DOS Path.			    */

int do_rezip(char *filename)
{
   char dos_call[64];
   if (system("md pkztmpxj") != 0)
       return 1;
   combine2(dos_call, filename);
   if (system(dos_call) != 0)
       return 1;
   combine3(dos_call, filename);
   if (system(dos_call) != 0)
       return 1;
   return 0;
}

/* This fucntion creates the DOS command line string to call PKUNZIP	    */

void combine2(char *str1, char *str2)
{
    strcpy(str1,"pkunzip -d ");
    strcat(str1,str2);
    strcat(str1," pkztmpxj");
}

/* This fucntion creates the DOS command line string to call PKZIP	    */

void combine3(char *str1, char *str2)
{
    strcpy(str1,"pkzip -ex -m -p -r ");
    strcat(str1,str2);
    strcat(str1," pkztmpxj\\*.*");
}
